

/******Gav's Generic Random Number Generators ****/
/*
    Based on ran2, from Numerical Recipies in C.
    
    Version 1.5
    Was broken.  GenRandomize igonores your seed. Whats wrong with this line?
        if (seed1 =1)  seed1=ALTSEED;
    Fixed 3/23/95.  Let this be a lesson to you.
    Added all sort of debug code, #define DEBUG

    Version 2.0 May 95
        Add several features
            genRanf now a macro
        
        2.2 Bug Fixing new stuff.
            Abandoned idea of having multiple generator types.
            easier to have completly different versions, and just recompile.
            GenRanSave and genRanRestore written. Needs insanity checking.
            Need to remove following line from genRan
                if(seed==0) genRandomize(0); (incase forget to call GenRandomize first)
            Can do so by initialising seed1, seed2, RanNum, and table[]
        3.0 Incorporate directly into 'gUtils'
            Remove genRan and only provide genRanf.  Its the only safe way
            of generating numbers.
            #define an inline fucntin genRan(int x) where x is the range
                produces numbers from 0 to (x-1)
        3.1 June 96
            return( (double) RanNum / ((double) GENRANMAX+1));
            Added the +1 to ensure that cannot return precisly 1.00000
        3.2 Jan 98
            Add code to generate a normally distributed random variable.
*/


#include <math.h>
#include <stdlib.h>
#include <time.h>
#include <limits.h>

#include "genRan.h"




/*****************************************************************
    The following code fragment came from <usenet://comp.lang.c.moderated>
    Uses unix specific calls to produce a random seed
    Kudos to danpop@cernapo.cern.ch (Dan Pop)
*****************************************************************/

long    genRanTime(void)
{
    return( (unsigned)clock() + (unsigned)time(NULL) );
}



/*******************LBD STARTS HERE *******************
This Generator is based on ran2 from numerical recipies in C,
2nd edition, p282.
L'Ecuyer generator with Bays-Durham shuffle.
Somewhat rewritten to make it somewhat more readable.
*******************************************************/

#define IM1 GENRANMAX
#define IM2 2147483399
#define IMM1 (IM1-1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32                 /* Table Size */
#define NDIV (1+IMM1/NTAB)

static long seed1=0;
static long seed2=0;
static long RanNum=0;
static long table[NTAB];

/*****************************************************************
    GenRandomize.  Initialises generator with seed via call
                    to specific GenRandomize;
            Returns 0 on sucess, 1 on non-fatal error
            Call with 0 for random(based on time) start.
*****************************************************************/

void  genRandomize(long seed)
{
    int     i,j;
    long    k;
    int     returnFlag=0;
    
    seed1=seed;
    if (seed1 < 0) seed1 =-(seed1);
    if (seed1 ==0) seed1 = genRanTime();
    
    seed2=(seed1);

    for(i=0;i<2;i++)            /*loop twice to 'warm up' the generator!?*/
        for (j=0;j<NTAB;j++) 
        {
            k=(seed1)/IQ1;
            seed1=IA1*(seed1-k*IQ1)-k*IR1;
            if (seed1 < 0) seed1 += IM1;
            table[j] = seed1;
        }
    RanNum=table[0];
}

int genRan(void)
{
    int j;
    long k;

    if(seed1==0) genRandomize(0); /*incase forget to call GenRandomize first*/
                    
    k=(seed1)/IQ1;
    seed1=IA1*(seed1-k*IQ1)-k*IR1;

    k=seed2/IQ2;
    seed2=IA2*(seed2-k*IQ2)-k*IR2;

    if (seed1 < 0) seed1 += IM1;
    if (seed2 < 0) seed2 += IM2;

    j=RanNum/NDIV;
    RanNum=table[j]-seed2;
    table[j] = seed1;
    if (RanNum < 1) RanNum += IMM1;
    
    return( RanNum );
}


double genRanf(void)
{
    return( (double) genRan() / ((double) GENRANMAX+1));
}

/******************************************************
        Return 32 random bits.
******************************************************/

unsigned int    genRan32(void)
{
    unsigned int    ran1;
    unsigned int    ran2;
    
    ran1= genRan();
    ran2= genRan();
    
    return( (ran1>>4)^(ran2<<8) );
}

/******************************************************
Return 32 bits, each bit set with the given probability
******************************************************/
/** Where did I get this from? ***/
unsigned    int genRanBits(double prob)
{
    #define ACCURACY    16
    #define MULT        65535
    #define ALLBITS     ULONG_MAX
    
    unsigned int    bitMask;
    unsigned int    probBits;
    unsigned int    ran;
    unsigned int    diff=0;
    unsigned int    mask = ALLBITS;
    unsigned int    out=0;
    
    int i;
    
    bitMask =(1<<16);

    probBits=(unsigned int) ( prob*MULT);
    /*printf("\n\n%X\t%X\n",probBits,mask);*/
    
    for(i=0;(mask !=0)&&(i<ACCURACY);i++)
    {
        bitMask =bitMask>>1;
        
        ran=genRan32();
        diff =((probBits & bitMask) ==0 ? ran : ~ran);
        out  = out |(diff & ran & mask);
        mask = mask&(~diff);

        /*printf("%X\t%X\t%X\t%X\t%X\n",bitMask,ran,diff,out,mask);*/
    }
    /*printf("%lg\t%d\n",prob,i);*/
    

    return ~out;    
}


        
    
/** Generate a normally distributed random number ***/
    
    
double  r1;
double  r2;
int flag=0;
double fac;

double  genRanGaussian()
{
    double  rsq=0.0;
    
    if(flag==0)
    {
        do
        {
            r1 = (genRanf()*2.)-1.;
            r2 = (genRanf()*2.)-1.;
            
            rsq = r1 * r1 + r2 * r2;
            
        } while( (rsq >1.0) || (rsq ==0.0) );
        
        
        fac =sqrt(-2.* log(rsq)/rsq);
    /*  printf("%lg\t%lg\t%lg\t%lg\n",r1,r2,rsq,fac);*/
        
        flag=1;
        return(r1*fac);
    } else {
        flag=0;
        return(r2*fac);
    }
    
}


